import logging
import random

from ._state import SeedState

_log = logging.getLogger(__name__)


def is_available():
    return True


def seed(state: SeedState):
    _log.debug("initializing stdlib seed")
    random.seed(state.int_seed)



import logging
import warnings

import numpy as np

from ._state import SeedState

try:
    from numba import njit
except ImportError:
    njit = None

_log = logging.getLogger(__name__)

if njit is not None:

    @njit  # type: ignore
    def _seed_numba(seed: int):
        np.random.seed(seed)


def is_available():
    return njit is not None


def seed(state: SeedState):
    if njit is None:
        warnings.warn("numba not available, skipping seed")
        return

    _log.debug("initializing Numba seed")
    _seed_numba(state.int_seed)



import os.path
from pathlib import Path

from seedbank import _root_state, init_file, root_seed

test_dir = os.path.dirname(__file__)


def test_init_toml_fn():
    "Initialize with a TOML filename"
    toml_file = os.path.join(test_dir, "init_seed.toml")
    init_file(toml_file)
    assert _root_state.seed.entropy == 202142
    assert len(_root_state.seed.spawn_key) == 0
    assert root_seed().entropy == 202142


def test_init_toml():
    "Initialize witha TOML path"
    toml_file = Path(test_dir) / "init_seed.toml"
    init_file(toml_file)
    assert _root_state.seed.entropy == 202142
    assert len(_root_state.seed.spawn_key) == 0
    assert root_seed().entropy == 202142


def test_init_JSON():
    "Initialize with a JSON path"
    toml_file = Path(test_dir) / "init_seed.json"
    init_file(toml_file)
    assert _root_state.seed.entropy == 202362
    assert len(_root_state.seed.spawn_key) == 0
    assert root_seed().entropy == 202362


def test_init_yaml():
    "Initialize with a YAML path"
    yaml_file = Path(test_dir) / "init_seed.yaml"
    init_file(yaml_file)
    assert _root_state.seed.entropy == 202157
    assert len(_root_state.seed.spawn_key) == 0
    assert root_seed().entropy == 202157


def test_init_toml_path():
    "Initialize with TOML and a custom configuration key."
    toml_file = Path(test_dir) / "init_seed.toml"
    init_file(toml_file, path="bird.seed")
    assert _root_state.seed.entropy == 202199
    assert len(_root_state.seed.spawn_key) == 0
    assert root_seed().entropy == 202199



"""
Tests for processing seed material.
"""

import numpy as np

from seedbank._keys import make_key


def test_make_key_int():
    key = make_key(42)
    assert key == 42


def test_make_key_npint():
    key = make_key(np.int32(42))
    assert key == 42


def test_make_key_str():
    key = make_key("hello, world")
    assert isinstance(key, np.ndarray)
    assert len(key) >= 4


def test_make_key_bytes():
    key = make_key(b"hello, world")
    assert isinstance(key, np.ndarray)
    assert len(key) >= 4


def test_make_key_ndarray():
    src = np.random.randint(500, size=32)
    key = make_key(src)
    assert isinstance(key, np.ndarray)
    assert all(key == src)


def test_make_key_str_int():
    key = make_key("hello, world", True)
    assert isinstance(key, int)



"""
Common infrastructure for initializing random number generators.
"""
# pyright: reportUnknownVariableType=false
from __future__ import annotations

import logging
from importlib import import_module
from importlib.metadata import PackageNotFoundError, version
from types import ModuleType

import numpy as np
from typing_extensions import Optional

from seedbank._keys import RNGKey, SeedLike, make_seed
from seedbank._state import SeedState

try:
    __version__ = version("seedbank")
except PackageNotFoundError:
    # package is not installed
    pass

_log = logging.getLogger(__name__)
_root_state = SeedState()

__all__ = [
    "make_seed",
    "initialize",
    "init_file",
    "derive_seed",
    "numpy_rng",
    "numpy_random_state",
    "cupy_rng",
]

# This list contains the modules that initialize seeds.
SEED_INITIALIZERS: list[str | ModuleType] = [
    "seedbank.stdlib",
    "seedbank.numpy",
    "seedbank.numba",
    "seedbank.cupy",
    "seedbank.tensorflow",
    "seedbank.torch",
]


def initialize(seed: SeedLike, *keys: RNGKey):
    """
    Initialize the random infrastructure with a seed.  This function should generally be
    called very early in the setup.  This initializes all known and available RNGs with
    a seed derived from the specified seed.

    If you do **not** call this function, a default root seed is used, so functions like
    :func:`derive_seed` and :func:`numpy_rng` work, but all other random number generators
    are left to their own default seeding behavior.

    Args:
        seed:
            The random seed to initialize with.
        keys:
            Additional keys, to use as a ``spawn_key`` on the seed sequence.
            Passed to :func:`derive_seed`.
    Returns:
        The random seed.
    """
    _root_state.initialize(seed, keys)
    _log.info("initialized root seed %s", _root_state.seed)

    for mod in SEED_INITIALIZERS:
        if isinstance(mod, str):
            mod = import_module(mod)
        if mod.is_available():
            mod.seed(_root_state)

    return _root_state.seed


def derive_seed(*keys: RNGKey, base: Optional[np.random.SeedSequence] = None):
    """
    Derive a seed from the root seed, optionally with additional seed keys.

    Args:
        keys:
            Additional components to add to the spawn key for reproducible derivation.
            If unspecified, the seed's internal counter is incremented (by calling
            :meth:`numpy.random.SeedSequence.spawn`).
        base:
            The base seed to use.  If ``None``, uses the root seed.

    Returns:
        The random seed.
    """
    return _root_state.derive(base, keys).seed


def root_seed():
    """
    Get the current root seed.

    Returns:
        numpy.random.SeedSequence:
            The root seed.
    """
    return _root_state.seed


def int_seed(
    words: Optional[int] = None, seed: Optional[np.random.SeedSequence] = None
) -> int | np.ndarray[int, np.dtype[np.uint32 | np.uint64]]:
    """
    Get the current root seed as an integer.

    Args:
        words:
            The number of words of entropy to return, or ``None`` for a single integer.
        seed:
            An alternate seed to convert to an ingeger; if ``None``, returns the root seed.

    Returns:
        The seed entropy.
    """

    if seed is None:
        seed = _root_state.seed

    if words is None:
        return seed.generate_state(1)[0]
    else:
        return seed.generate_state(words)


from seedbank._config import init_file  # noqa: E402
from seedbank.cupy import cupy_rng  # noqa: E402
from seedbank.numpy import numpy_random_state, numpy_rng  # noqa: E402



# type: ignore
import logging
import warnings

try:
    import torch
except ImportError:
    torch = None

_log = logging.getLogger(__name__)


def is_available():
    return torch is not None


def seed(state):
    if torch is None:
        warnings.warn("torch not available, skipping seed")
        return

    seed = state.int_seed
    torch.manual_seed(seed)



from pytest import mark
import numpy as np

try:
    import cupy
except ImportError:
    pytestmark = mark.skip("cupy not available")

from seedbank import cupy_rng


def test_generator():
    rng = cupy_rng()
    assert isinstance(rng, cupy.random.Generator)


def test_generator_seed():
    rng = cupy_rng(42)
    assert isinstance(rng, cupy.random.Generator)


def test_generator_seed_seq():
    seq = np.random.SeedSequence(42)
    rng = cupy_rng(seq)
    assert isinstance(rng, cupy.random.Generator)


def test_generator_passthrough():
    rng1 = cupy.random.default_rng()
    rng = cupy_rng(rng1)
    assert isinstance(rng, cupy.random.Generator)
    assert rng is rng1



# pyright: reportUnnecessaryIsInstance=false
import hashlib
from typing import Any, Literal, Sequence, TypeAlias, overload

import numpy as np
import numpy.typing as npt

Entropy: TypeAlias = int | Sequence[int] | npt.NDArray[np.uint32]
RNGKey: TypeAlias = int | np.integer[Any] | npt.NDArray[Any] | bytes | memoryview | str
SeedLike: TypeAlias = np.random.SeedSequence | RNGKey


@overload
def make_key(data: RNGKey, single: Literal[True]) -> int:
    ...


@overload
def make_key(data: RNGKey, single: bool = False) -> Entropy:
    ...


def make_key(data: RNGKey, single: bool = False) -> Entropy:
    """
    Get a key, usable as entropy in a seed sequence, from a piece of data.

    Args:
        data: The key data.
        single: If ``True``, always return a single integer.
    """
    if isinstance(data, int) or isinstance(data, np.integer):
        return int(data)
    if isinstance(data, np.ndarray):
        if single:
            return make_key(memoryview(data), True)
        else:
            return data.astype(np.uint32)
    if isinstance(data, (bytes, memoryview)):
        size = 8 if single else 32
        h = hashlib.blake2b(data, digest_size=size)
        seed = np.frombuffer(h.digest(), np.uint64)
        if single:
            return int(seed[0])
        else:
            return seed.astype(np.uint32)
    if isinstance(data, str):
        return make_key(data.encode("utf8"), single)

    # never reached for type-checked code but we want to be robust
    dt = type(data)
    raise TypeError(f"invalid seed type {dt}")


def make_seed(data: SeedLike) -> np.random.SeedSequence:
    """
    Get a seed sequence from a piece of data.

    Args:
        data: The seed material.

    Returns:
        A seed sequence suitable for initializing RNGs.
    """
    if isinstance(data, np.random.SeedSequence):
        return data

    entropy = make_key(data)
    return np.random.SeedSequence(entropy)



import logging

import numpy as np
from typing_extensions import Optional, TypeAlias

from . import derive_seed, int_seed
from ._keys import SeedLike, make_seed
from ._state import SeedState

_log = logging.getLogger(__name__)

NPRNGSource: TypeAlias = SeedLike | np.random.Generator | np.random.RandomState


def is_available():
    return True


def seed(state: SeedState):
    _log.debug("initializing NumPy root RNG")
    np.random.seed(state.int_seed)


def numpy_rng(
    spec: Optional[NPRNGSource] = None,
) -> np.random.Generator:
    """
    Get a NumPy random number generator.  This is similar to
    :func:`sklearn.utils.check_random_state`, but it returns a
    :class:`~numpy.random.Generator` instead.

    Args:
        spec:
            The spec for this RNG. Can be any of the following types:

            * :data:`SeedLike`
            * :class:`numpy.random.Generator` — returned as-is
            * :class:`numpy.random.RandomState` — its ``_bit_generator`` is
              extracted and wrapped in a :class:`~numpy.random.Generator`.

    Returns:
        A random number generator.
    """

    if isinstance(spec, np.random.Generator):
        return spec
    elif isinstance(spec, np.random.RandomState):
        return np.random.Generator(spec._bit_generator)
    elif spec is None:
        return np.random.default_rng(derive_seed())
    else:
        seed = make_seed(spec)
        return np.random.default_rng(seed)


def numpy_random_state(spec: Optional[NPRNGSource] = None) -> np.random.RandomState:
    """
    Get a legacy NumPy random number generator
    (:class:`~numpy.random.mtrand.RandomState`). This is similar to
    :func:`sklearn.utils.check_random_state`.

    Args:
        spec:
            The spec for this RNG.  Can be any of the following types:

            * :data:`SeedLike`
            * :class:`numpy.random.RandomState` — returned as-is
            * :class:`numpy.random.Generator` — its
              :attr:`~numpy.random.Generator.bit_generator` is extracted and
              wrapped in a :class:`~numpy.random.RandomState`.

    Returns:
        A random number generator.
    """

    if isinstance(spec, np.random.RandomState):
        return spec
    elif isinstance(spec, np.random.Generator):
        return np.random.RandomState(spec.bit_generator)
    elif spec is None:
        return np.random.RandomState(int_seed(seed=derive_seed()))
    else:
        seed = make_seed(spec)
        return np.random.RandomState(int_seed(seed=seed))



# type: ignore
from __future__ import annotations

import logging
import warnings

from typing_extensions import Optional

from . import derive_seed
from ._keys import SeedLike, make_seed

try:
    import cupy
except ImportError:
    cupy = None

_log = logging.getLogger(__name__)


def is_available():
    return cupy is not None


def seed(state):
    if cupy is None:
        warnings.warn("cupy not available, skipping seed")

    _log.debug("initializing CuPy root RNG")
    cupy.random.seed(state.int_seed)


def cupy_rng(spec: Optional[SeedLike | cupy.random.Generator] = None):
    """
    Get a CuPy random number generator.  This works like :func:`numpy_rng`, but
    it returns a :class:`cupy.random.Generator` instead.

    Args:
        spec:
            The spec for this RNG.  Can be any of the following types:

            * ``int``
            * ``None``
            * :class:`numpy.random.SeedSequence`
            * :class:`numpy.random.RandomState` (its bit-generator is extracted
              and wrapped in a generator)
            * :class:`numpy.random.Generator` (returned as-is)

    Returns:
        cupy.random.Generator: A random number generator.
    """
    import cupy

    if isinstance(spec, cupy.random.Generator):
        return spec
    elif spec is None:
        return cupy.random.default_rng(derive_seed())
    else:
        seed = make_seed(spec)
        return cupy.random.default_rng(seed)



"""
Configuration file parsing support.
"""
from __future__ import annotations

import logging
from os import PathLike
from pathlib import Path
from typing import Any

from . import initialize
from ._keys import RNGKey

_log = logging.getLogger("seedbank.config")


def _parse_toml(file: Path) -> dict[str, Any]:
    try:
        from tomllib import loads
    except ImportError:
        from toml import loads

    _log.debug("parsing TOML from {}", file)
    return loads(file.read_text())


def _parse_yaml(file: Path) -> dict[str, Any]:
    from yaml import SafeLoader, load

    _log.debug("parsing YAML from {}", file)
    with file.open("r") as f:
        return load(f, SafeLoader)


def _parse_json(file: Path) -> dict[str, Any]:
    from json import loads

    _log.debug("parsing JSON from {}", file)
    return loads(file.read_text())


def init_file(file: str | PathLike[str], *keys: RNGKey, path: str = "random.seed"):
    """
    Initialize the random infrastructure with a seed loaded from a file. The loaded seed is
    passed to :func:`initialize`, along with any additional RNG key material.

    With the default ``path``, the seed can be configured from a TOML file as follows:

    .. code-block:: toml

        [random]
        seed = 2308410

    And then initialized::

        seedbank.init_file('params.toml')

    Any file type supported by anyconfig_ can be used, including TOML, YAML, and JSON.

    .. _anyconfig: https://github.com/ssato/python-anyconfig

    Args:
        file:
            The filename for the configuration file to load.
        keys:
            Aditional key material.
        path:
            The path within the configuration file or object in which the seed is stored.
            Can be multiple keys separated with '.'.
    """

    _log.info("loading seed from %s (key=%s)", file, path)
    file = Path(file)
    match file.suffix.lower():
        case ".toml":
            parser = _parse_toml
        case ".json":
            parser = _parse_json
        case ".yml" | ".yaml":
            parser = _parse_yaml
        case _:
            raise ValueError(f"unsupported file type {file.suffix}")

    config = parser(file)

    kps = path.split(".")
    cvar = config
    for k in kps:
        cvar = cvar[k]  # type: ignore

    initialize(cvar, *keys)  # type: ignore
